####################
# File name: A5.py
# This is the programming portion of Assignment 5
# for CPSC 447/547 Intro to Quantum Computing
#
# Name: 
# NetID:
####################

import requirement_A5
import numpy as np
# run `pip3 install fusion-blossom` to install
import fusion_blossom as fb

####################
# Place for your own helper functions, if any
####################



####################
# Building Blocks
####################

class Qubit(object):
    """Qubit object"""
    def __init__(self, arg, label='q'):
        super(Qubit, self).__init__()
        self.arg = arg
        self.label = label

class QuantumRegister(object):
    """QuantumRegister is where we keep track of qubits"""
    def __init__(self, num_q, label='qreg'):
        super(QuantumRegister, self).__init__()
        self.size = num_q
        self.label = label
        self.array = [Qubit(i) for i in range(num_q)]
        self.pauli_x_errors = [False for _ in range(num_q)]

class ClassicalRegister(object):
    """ClassicalRegister is where we keep track of measurement outcomes"""
    def __init__(self, num_c, label='creg'):
        super(ClassicalRegister, self).__init__()
        self.size = num_c
        self.label = label
        self.state = [bool(0) for _ in range(num_c)]

class Gate(object):
    """Gate object to describe its name, kind"""
    def __init__(self, name, num_q, matrix):
        super(Gate, self).__init__()
        self.name = name 
        self.num_q = num_q
        assert matrix is None, ""
    def __repr__(self):
        return self.name

    def commute_through(self, pauli_x_errors_before):
        pauli_x_errors_after = [is_error for is_error in pauli_x_errors_before]
        if self.name == "x_error":
            assert len(pauli_x_errors_before) == 1
            # YOUR IMPLEMENTATION HERE
            
        elif self.name == "cx":
            assert len(pauli_x_errors_before) == 2
            # YOUR IMPLEMENTATION HERE
            
        elif self.name == "initialize":
            # YOUR IMPLEMENTATION HERE

            pass
        return pauli_x_errors_after

####################
# In-House Quantum Circuit
####################

class QuantumCircuit(object):
    """QuantumCircuit"""
    def __init__(self, num_q, num_c):
        super(QuantumCircuit, self).__init__()
        self.num_q = num_q
        self.qubits = QuantumRegister(num_q) # initialized qubits
        self.num_c = num_c
        self.cbits = ClassicalRegister(num_c) # initialized cbits
        self.circuit = [] # sequence of instructions
        self.pc = 0 # program counter

    def _append(self, operation, q_array, c_array):
        # Add new instruction to circuit
        instruction = [operation, q_array, c_array]
        self.circuit.append(instruction)

    def __repr__(self):
        # For displaying quantum circuit
        qasm = ['\n======<CPSC 447/547 QASM>======']
        qasm += ['Qreg: %d, Creg: %d' % (self.num_q, self.num_c)]
        for inst in self.circuit:
            #print(inst)
            (op, q_arr, c_arr) = inst
            inst_str = '%s ' % op.name 
            for q in q_arr:
                qubit = self.qubits.array[q]
                inst_str += '%s%d ' % (qubit.label, qubit.arg)
            inst_str += ', '
            for c in c_arr:
                inst_str += '%s%d ' % (self.cbits.label, c)
            qasm.append(inst_str)
        qasm.append('===============================\n')
        return "\n".join(qasm)

    # Define instruction set
    # Pauli X (error) gate
    def x_error(self, qubit):
        # Define Gate by its name, kind (number of qubit), and matrix
        XErrorGate = Gate('x_error', 1, None)
        # add to the end of the self.circuit list
        self._append(XErrorGate, [qubit], [])
        return
    # Controlled X gate (CNOT)
    def cx(self, ctrl_qubit, trgt_qubit):
        CXGate = Gate('cx', 2, None)
        self._append(CXGate, [ctrl_qubit, trgt_qubit], [])
        return
    # Measure qubits in array 'qubits' and store classical outcome in 'cbits'
    def measure(self, qubits, cbits):
        assert(len(qubits) == len(cbits))
        Measure = Gate('measure', len(qubits), None)
        self._append(Measure, qubits, cbits)
        return
    # (Re)initialize qubits into |0> state
    def initialize(self, qubits):
        Initialize = Gate('initialize', len(qubits), None)
        self._append(Initialize, qubits, [])
        return

    ####################
    # In-House Pauli Error Simulator
    ####################

    def evolveOneStep(self):
        # Evolve one step from program counter pc, update pauli x errors vector
        # Feel free to use/modify this helper function for simulate.
        # OUTPUT: - Return quantum state after one step of evolution
        #         - Update self.curr_state to new quantum state
        #         - Increment program counter pc
        #         - Return None if evolving measurement
        # ASSUME: - The initial state is the +1 eigenstate of all measurement operators
        #         - Measurements can be in the circuit, following the measurement rule
        (op, q_arr, c_arr) = self.circuit[self.pc]
        self.pc += 1
        if op.name != 'measure':
            # YOUR IMPLEMENTATION HERE
            
            return self.qubits.pauli_x_errors
        else:
            assert len(q_arr) == len(c_arr)
            # YOUR IMPLEMENTATION HERE
            
            return None

    def simulate(self):
        # A function for simulating pauli errors from scratch.
        # OUTPUT: - Return both the measurement results and the Pauli X error vector
        # ASSUME: - The initial state is the +1 eigenstate of all measurement operators
        #         - Measurements can be in the circuit, following the measurement rule
        while self.pc < len(self.circuit):
            self.evolveOneStep()
        return self.cbits.state, self.qubits.pauli_x_errors

####################
# In-House Quantum Repetition Code and Decoder
####################

class QuantumRepetitionCode(object):
    """Quantum Repetition Code Circuit Builder"""
    def __init__(self, d, noisy_measurements = 0):
        # d: code distance
        # noisy_measurements rounds of measurements + a final perfect measurement
        # qubits layout: [data, ancilla, data, ancilla, data, ..., ancilla, data]
        super(QuantumRepetitionCode, self).__init__()
        self.d = d
        self.noisy_measurements = noisy_measurements
        self.num_c = (self.d-1) * (self.noisy_measurements+1)
        self.num_q = self.d*2-1
        self.simultaneous_gates = []  # apply multiple gates simultaneously
        for round in range(noisy_measurements):
            self.simultaneous_gates.append([])  # empty slot for errors (step 0)
            self.simultaneous_gates.append([('initialize', [2*i+1 for i in range(d-1)])])
            self.simultaneous_gates.append([])  # empty slot for errors (step 1)
            self.simultaneous_gates.append([('cx', 2*i, 2*i+1) for i in range(d-1)])
            self.simultaneous_gates.append([])  # empty slot for errors (step 2)
            self.simultaneous_gates.append([('cx', 2*i+2, 2*i+1) for i in range(d-1)])
            self.simultaneous_gates.append([])  # empty slot for errors (step 3)
            self.simultaneous_gates.append([('measure', [2*i+1 for i in range(d-1)]
                , [i+round*(d-1) for i in range(d-1)])])
        self.simultaneous_gates.append([])  # empty slot for errors (`noisy_measurements`, 0)
        self.simultaneous_gates.append([('initialize', [2*i+1 for i in range(d-1)])])
        self.simultaneous_gates.append([('cx', 2*i, 2*i+1) for i in range(d-1)])
        self.simultaneous_gates.append([('cx', 2*i+2, 2*i+1) for i in range(d-1)])
        self.simultaneous_gates.append([('measure', [2*i+1 for i in range(d-1)]
            , [i+noisy_measurements*(d-1) for i in range(d-1)])])

    # Manually add x errors to some qubits at specific round and step
    def add_x_errors(self, round, step, q_arr):
        assert round >= 0 and (round < self.noisy_measurements or (round == self.noisy_measurements and step == 0))
        assert step >= 0 and step < 4
        gates = self.simultaneous_gates[round*8 + step*2]
        for q_idx in q_arr:
            gates.append(('x_error', q_idx))

    # create a QuantumCircuit including all the operations
    def circuit(self):
        qc = QuantumCircuit(self.num_q, self.num_c)
        for gates in self.simultaneous_gates:
            for gate in gates:
                gate_name = gate[0]
                gate_args = gate[1:]
                func = getattr(qc, gate_name)
                func(*gate_args)
        return qc

    def clear_noise(self):
        for round in range(self.noisy_measurements):
            for step in range(4):
                self.simultaneous_gates[round*8 + step*2].clear()
        self.simultaneous_gates[self.noisy_measurements*8].clear()

    def decode_code_capacity_noise(self, p, measurements, visualize=False):
        assert self.noisy_measurements == 0, "code capacity noise model assumes no noisy measurement"
        corrections = []
        vertex_num = (self.d - 1) + 2
        virtual_vertices = [0, vertex_num - 1]
        weighted_edges = []

        # YOUR IMPLEMENTATION HERE: build weighted edges
        

        weighted_edges = [(u, v, 2 * int(max(we, 1))) for (u, v, we) in weighted_edges]  # make sure weights are even integer
        initializer = fb.SolverInitializer(vertex_num, weighted_edges, virtual_vertices)
        solver = fb.SolverSerial(initializer)
        positions = [fb.VisualizePosition(0, i, 0) for i in range(vertex_num)]
        visualize_filename = fb.static_visualize_data_filename()
        for measurement in measurements:
            assert len(measurement) == self.num_c
            correction = [False for _ in range(self.d)]  # correction only applies to data qubit
            visualizer = None
            if visualize:
                visualizer = fb.Visualizer(filepath=visualize_filename, positions=positions)  # comment out to disable
            defect_vertices = []

            # YOUR IMPLEMENTATION HERE: prepare defect vertices
            

            syndrome = fb.SyndromePattern(defect_vertices=defect_vertices)
            solver.solve(syndrome)
            subgraph = solver.subgraph(visualizer)
            if visualizer is not None:
                fb.print_visualize_link(filename=visualize_filename)
                fb.helper.open_visualizer(visualize_filename, open_browser=True)

            # build correction
            for edge_index in subgraph:
                # YOUR IMPLEMENTATION HERE
                
                pass

            solver.clear()
            corrections.append(correction)
        return corrections

    # apply phenomenological noise model (only data qubit error and measurement error)
    def apply_phenomenological_noise(self, p):
        for round in range(self.noisy_measurements):
            for i in range(self.d):
                if np.random.random_sample() < p:
                    self.add_x_errors(round, 0, [2 * i])
            for i in range(self.d - 1):
                if np.random.random_sample() < p:
                    self.add_x_errors(round, 3, [2 * i + 1])
        for i in range(self.d):
            if np.random.random_sample() < p:
                self.add_x_errors(self.noisy_measurements, 0, [2 * i])

    def decode_phenomenological_noise(self, p, measurements, visualize=False):
        if self.noisy_measurements == 0:
            return self.decode_code_capacity_noise(p, measurements)
        else:
            corrections = []
            row_num = (self.d - 1) + 2
            vertex_num = row_num * (self.noisy_measurements + 1)
            virtual_vertices = [row_num*round for round in range(self.noisy_measurements + 1)]
            virtual_vertices += [row_num*(round+1)-1 for round in range(self.noisy_measurements + 1)]
            weighted_edges = []

            # YOUR IMPLEMENTATION HERE: build weighted edges
            

            weighted_edges = [(u, v, 2 * int(max(we, 1))) for (u, v, we) in weighted_edges]  # make sure weights are even integer
            initializer = fb.SolverInitializer(vertex_num, weighted_edges, virtual_vertices)
            solver = fb.SolverSerial(initializer)
            positions = []
            for round in range(self.noisy_measurements + 1):
                for i in range(row_num):
                    positions.append(fb.VisualizePosition(-round, i, 0))
            visualize_filename = fb.static_visualize_data_filename()
            for measurement in measurements:
                assert len(measurement) == self.num_c
                correction = [False for _ in range(self.d)]  # correction only applies to data qubit
                visualizer = None
                if visualize:
                    visualizer = fb.Visualizer(filepath=visualize_filename, positions=positions)  # comment out to disable
                defect_vertices = []

                # YOUR IMPLEMENTATION HERE: prepare defect vertices
                

                syndrome = fb.SyndromePattern(defect_vertices=defect_vertices)
                solver.solve(syndrome)
                subgraph = solver.subgraph(visualizer)
                if visualizer is not None:
                    fb.print_visualize_link(filename=visualize_filename)
                    fb.helper.open_visualizer(visualize_filename, open_browser=True)

                # build correction
                for edge_index in subgraph:
                    # YOUR IMPLEMENTATION HERE
                    
                    pass

                solver.clear()
                corrections.append(correction)
            return corrections

    # apply circuit-level noise (error possible anywhere)
    def apply_circuit_level_noise(self, p):
        for round in range(self.noisy_measurements):
            for step in range(4):
                for i in range(self.num_q):
                    if np.random.random_sample() < p:
                        self.add_x_errors(round, step, [i])
        for i in range(self.num_q):
            if np.random.random_sample() < p:
                self.add_x_errors(self.noisy_measurements, 0, [i])

    def decode_circuit_level_noise(self, p, measurements, visualize=False):
        if self.noisy_measurements == 0:
            return self.decode_code_capacity_noise(p, measurements)
        else:
            corrections = []
            row_num = (self.d - 1) + 2
            vertex_num = row_num * (self.noisy_measurements + 1)
            virtual_vertices = [row_num*round for round in range(self.noisy_measurements + 1)]
            virtual_vertices += [row_num*(round+1)-1 for round in range(self.noisy_measurements + 1)]
            weighted_edges = []

            # YOUR IMPLEMENTATION HERE: build weighted edges
            

            weighted_edges = [(u, v, 2 * int(max(we, 1))) for (u, v, we) in weighted_edges]  # make sure weights are even integer
            initializer = fb.SolverInitializer(vertex_num, weighted_edges, virtual_vertices)
            solver = fb.SolverSerial(initializer)
            positions = []
            for round in range(self.noisy_measurements + 1):
                for i in range(row_num):
                    positions.append(fb.VisualizePosition(-round, i, 0))
            visualize_filename = fb.static_visualize_data_filename()
            for measurement in measurements:
                assert len(measurement) == self.num_c
                correction = [False for _ in range(self.d)]  # correction only applies to data qubit
                visualizer = None
                if visualize:
                    visualizer = fb.Visualizer(filepath=visualize_filename, positions=positions)  # comment out to disable
                defect_vertices = []

                # YOUR IMPLEMENTATION HERE: prepare defect vertices
                

                syndrome = fb.SyndromePattern(defect_vertices=defect_vertices)
                solver.solve(syndrome)
                subgraph = solver.subgraph(visualizer)
                if visualizer is not None:
                    fb.print_visualize_link(filename=visualize_filename)
                    fb.helper.open_visualizer(visualize_filename, open_browser=True)

                # build correction
                for edge_index in subgraph:
                    # YOUR IMPLEMENTATION HERE
                    
                    pass

                solver.clear()
                corrections.append(correction)
            return corrections

    def __repr__(self):
        qcirc = ['\n======<CPSC 447/547 Quantum Circuit>======']
        qcirc += ['Qreg: %d, Creg: %d' % (self.num_q, self.num_c)]
        lines = ["" for i in range(self.num_q * 3)]
        for q_idx in range(self.num_q):
            name = f"data" if q_idx % 2 == 0 else "stab"
            lines[q_idx * 3] += " " * 8
            lines[q_idx * 3 + 1] += f"{name} {q_idx:2d} ─"
            lines[q_idx * 3 + 2] += " " * 8
        blank = lambda li: ("─" if li % 3 == 1 else " ")
        for gates_idx, gates in enumerate(self.simultaneous_gates):
            for gate in gates:
                gate_name = gate[0]
                if gate_name == "initialize":
                    for qubit in gate[1]:
                        max_depth = max([len(lines[li]) for li in range(qubit * 3, (qubit + 1) * 3)])
                        for li in range(qubit * 3, (qubit + 1) * 3):
                            lines[li] += blank(li) * (max_depth - len(lines[li]))
                        lines[3*qubit + 0] += "┌───┐"
                        lines[3*qubit + 1] += "┤|0>├"
                        lines[3*qubit + 2] += "└───┘"
                elif gate_name == "cx":
                    ctrl_qubit, trgt_qubit = gate[1:3]
                    lower_line = min(ctrl_qubit, trgt_qubit) * 3
                    upper_line = (max(ctrl_qubit, trgt_qubit) + 1) * 3
                    max_depth = max([len(lines[li]) for li in range(lower_line, upper_line)])
                    for li in range(lower_line, upper_line):
                        lines[li] += blank(li) * (max_depth - len(lines[li]))
                    lines[3*ctrl_qubit + 1] += "──■──"
                    lines[3*trgt_qubit + 1] += "┤ X ├"
                    if ctrl_qubit < trgt_qubit:
                        lines[3*ctrl_qubit + 0] += "     "
                        lines[3*ctrl_qubit + 2] += "  |  "
                        lines[3*trgt_qubit + 0] += "┌─┴─┐"
                        lines[3*trgt_qubit + 2] += "└───┘"
                    else:
                        lines[3*trgt_qubit + 0] += "┌───┐"
                        lines[3*trgt_qubit + 2] += "└─┬─┘"
                        lines[3*ctrl_qubit + 0] += "  |  "
                        lines[3*ctrl_qubit + 2] += "     "
                    for li in range(lower_line + 3, upper_line - 3):
                        lines[li] += blank(li) * 2 + "|" + blank(li) * 2
                elif gate_name == "x_error":
                    qubit = gate[1]
                    max_depth = max([len(lines[li]) for li in range(qubit * 3, (qubit + 1) * 3)])
                    for li in range(qubit * 3, (qubit + 1) * 3):
                        lines[li] += blank(li) * (max_depth - len(lines[li]))
                    lines[3*qubit + 0] += "┌╍╍╍┐"
                    lines[3*qubit + 1] += "┤EX ├"
                    lines[3*qubit + 2] += "└╍╍╍┘"
                elif gate_name == "measure":
                    qubits = gate[1]
                    cbits = gate[2]
                    for idx, qubit in enumerate(qubits):
                        cbit = cbits[idx]
                        max_depth = max([len(lines[li]) for li in range(qubit * 3, (qubit + 1) * 3)])
                        for li in range(qubit * 3, (qubit + 1) * 3):
                            lines[li] += blank(li) * (max_depth - len(lines[li]))
                        lines[3*qubit + 0] += "┌───┐"
                        lines[3*qubit + 1] += f"┤M{cbit:2d}├"
                        lines[3*qubit + 2] += "└─╥─┘"
            max_depth = max([len(line) for line in lines])
            for li in range(len(lines)):
                lines[li] += blank(li) * (max_depth - len(lines[li]))
            if gates_idx % 8 == 7:
                for li in range(len(lines)):
                    lines[li] += "┊"
        for li, line in enumerate(lines):
            qcirc.append(line + blank(li))
        qcirc.append('==========================================\n')
        return "\n".join(qcirc)

# evaluate logical error rate with fixed number of simulations
def evaluate_logical_error_rate(p, d, noisy_measurements, repeats=1000, is_circuit_level=False
        , force_phenomenological_decoder=False, debug_visualize_logical_error=False):
    code = QuantumRepetitionCode(d, noisy_measurements)
    measurement_vec = []
    pauli_x_errors_vec = []
    for i in range(repeats):
        if is_circuit_level:
            code.apply_circuit_level_noise(p)
        else:
            code.apply_phenomenological_noise(p)
        qc = code.circuit()
        measurement, pauli_x_errors = qc.simulate()
        measurement_vec.append(measurement)
        pauli_x_errors_vec.append(pauli_x_errors)
        code.clear_noise()
    decoder = QuantumRepetitionCode(d, noisy_measurements)
    if is_circuit_level and not force_phenomenological_decoder:
        corrections = decoder.decode_circuit_level_noise(p, measurement_vec)
    else:
        corrections = decoder.decode_phenomenological_noise(p, measurement_vec)
    logical_error_count = 0
    for i, correction in enumerate(corrections):
        if correction != pauli_x_errors_vec[i][0::2]:
            logical_error_count += 1
            if debug_visualize_logical_error:
                print(measurement_vec[i], pauli_x_errors_vec[i])
                decoder.decode_circuit_level_noise(p, [measurement_vec[i]], visualize=True)
                exit(0)
    error_rate = logical_error_count / repeats
    # calculate relative uncertainty of `error_rate` with 95% confidence interval
    relative_uncertainty = float("NAN")
    if error_rate > 0:
        relative_uncertainty = 1.96 * np.sqrt(error_rate * (1. - error_rate) / repeats) / error_rate
    return error_rate, relative_uncertainty

####################
# Test functions (feel free to add more of your own tests here)
####################

def testCommuteThroughCXGeneric(pauli_x_errors_before, expected_pauli_x_errors_after):
    CXGate = Gate('cx', 2, None)
    pauli_x_errors_after = CXGate.commute_through(pauli_x_errors_before)
    assert pauli_x_errors_after == expected_pauli_x_errors_after
def testCommuteThroughCX00():  # II -> II
    testCommuteThroughCXGeneric([False, False], [False, False])
def testCommuteThroughCX01():  # IX -> IX
    testCommuteThroughCXGeneric([False, True], [False, True])
def testCommuteThroughCX10():  # XI -> XX
    testCommuteThroughCXGeneric([True, False], [True, True])
def testCommuteThroughCX11():  # XX -> XI
    testCommuteThroughCXGeneric([True, True], [True, False])

def testCommuteThroughInitialize():
    InitializeGate = Gate('initialize', 2, None)
    pauli_x_errors_after = InitializeGate.commute_through([False, True])
    assert pauli_x_errors_after == [False, False]

def testBasicSimulation():
    # qubits layout: [data, ancilla, data, ancilla, data]
    code = QuantumRepetitionCode(3)
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    assert measurement == [False, False]  # of ancilla qubits
    assert pauli_x_errors == [False, False, False, False, False]  # of all qubits

def testNoisySimulation():
    # qubits layout: [data, ancilla, data, ancilla, data]
    code = QuantumRepetitionCode(3)
    code.add_x_errors(0, 0, [0])
    print(code)
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    assert measurement == [True, False]  # of ancilla qubits
    assert pauli_x_errors == [True, True, False, False, False]  # of all qubits

def testNoisySimulationMultipleRounds():
    # qubits layout: [data, ancilla, data, ancilla, data]
    code = QuantumRepetitionCode(3, noisy_measurements=2)
    code.add_x_errors(0, 2, [2])
    print(code)
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    assert measurement == [True, False, True, True, True, True]  # of ancilla qubits
    assert pauli_x_errors == [False, True, True, True, False]  # of all qubits

def testNoisySimulationAllPhenomenologicalNoise():
    # qubits layout: [data, ancilla, data, ancilla, data]
    code = QuantumRepetitionCode(3, noisy_measurements=2)
    np.random.seed(447547)
    code.apply_phenomenological_noise(p=1.0)
    print("\n\n[hint] below: all possible error position in phenomenological noise model")
    print(code)
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    assert measurement == [True, True, True, True, False, False]  # of ancilla qubits
    assert pauli_x_errors == [True, False, True, False, True]  # of all qubits

def testNoisySimulationAllCircuitLevelNoise():
    # qubits layout: [data, ancilla, data, ancilla, data]
    code = QuantumRepetitionCode(3, noisy_measurements=2)
    np.random.seed(447547)
    code.apply_circuit_level_noise(p=1.0)
    print("\n\n[hint] below: all possible error position in circuit-level noise model")
    print(code)
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    assert measurement == [False, False, False, False, False, False]  # of ancilla qubits
    assert pauli_x_errors == [True, False, True, False, True]  # of all qubits

def testBasicDecode():
    code = QuantumRepetitionCode(3)
    code.add_x_errors(0, 0, [0])
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    decoder = QuantumRepetitionCode(3)  # a new instance with same parameter
    corrections = decoder.decode_code_capacity_noise(0.01, [measurement], visualize=False)
    assert corrections[0] == pauli_x_errors[0::2]

def testDecodePhenomenological():
    code = QuantumRepetitionCode(3, 2)
    code.add_x_errors(0, 0, [0])
    code.add_x_errors(2, 0, [4])
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    decoder = QuantumRepetitionCode(3, 2)  # a new instance with same parameter
    corrections = decoder.decode_phenomenological_noise(0.01, [measurement], visualize=False)
    assert corrections[0] == pauli_x_errors[0::2]

def testDecodeCircuitLevel():
    code = QuantumRepetitionCode(3, 2)
    code.add_x_errors(0, 2, [2])
    qc = code.circuit()
    measurement, pauli_x_errors = qc.simulate()
    decoder = QuantumRepetitionCode(3, 2)  # a new instance with same parameter
    corrections = decoder.decode_circuit_level_noise(0.01, [measurement], visualize=False)
    assert corrections[0] == pauli_x_errors[0::2]

def testDecodeLogicalErrorRate():
    np.random.seed(235321)
    repeats = 1000
    p = 0.3
    d = 5
    error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, 0, repeats)
    print(f"code-capacity physical error rate: {p}, logical error rate: {error_rate} ± {error_rate * relative_uncertainty:.3f}")
    expected_error_rate = 0.162
    assert error_rate >= expected_error_rate / 1.5 and error_rate <= expected_error_rate * 1.5

def testDecodeLogicalErrorRatePhenomenological():
    np.random.seed(235321)
    repeats = 1000
    p = 0.05
    d = 5
    error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, d, repeats)
    print(f"phenomenological physical error rate: {p}, logical error rate: {error_rate} ± {error_rate * relative_uncertainty:.3f}")
    expected_error_rate = 0.0329
    assert error_rate >= expected_error_rate / 2.0 and error_rate <= expected_error_rate * 2.0

def testDecodeLogicalErrorRateCircuitLevel():
    np.random.seed(235321)
    repeats = 1000
    p = 0.01
    d = 5
    error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, d, repeats, is_circuit_level=True)
    print(f"circuit-level physical error rate: {p}, logical error rate: {error_rate} ± {error_rate * relative_uncertainty:.3f}")
    expected_error_rate = 0.016
    assert error_rate >= expected_error_rate / 2.0 and error_rate <= expected_error_rate * 2.0

# https://stackoverflow.com/questions/3160699/python-progress-bar
def progressbar(it, prefix="", size=40):
    count = len(it)
    def show(j):
        x = int(size*j/count)
        print(f"{prefix}[{u'█'*x}{('.'*(size-x))}] {j}/{count}", end='\r', flush=True)
    show(0)
    for i, item in enumerate(it):
        yield item
        show(i+1)
    print("\n", flush=True)

def evaluateDecoderThreshold():
    np.random.seed(3422712)
    repeats = 10000
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.ylim(2e-3, 0.55)
    plt.xlim(1e-2, 0.5)
    plt.yscale('log')
    plt.xscale('log')
    plt.xlabel('physical error rate')
    plt.ylabel('logical error rate')
    plt.title('Code-Capacity Noise Model')
    for d in [3, 5, 7, 9, 11]:
        p_vec = [0.5 * (0.8 ** i) for i in range(15)]
        y_vec = []
        y_error = []
        for p in progressbar(p_vec, f"d = {d:2d}: "):
            error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, 0, repeats)
            if relative_uncertainty > 0.3:
                break
            y_vec.append(error_rate)
            y_error.append(error_rate * relative_uncertainty)
        plt.errorbar(p_vec[:len(y_vec)], y_vec, y_error, label=f"d = {d}")
        print()
    plt.legend()
    # plt.show()
    plt.savefig('evaluateDecoderThreshold.pdf')

def evaluateDecoderThresholdPhenomenological():
    np.random.seed(3422712)
    repeats = 10000
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.ylim(2e-3, 0.55)
    plt.xlim(1e-2, 0.5)
    plt.yscale('log')
    plt.xscale('log')
    plt.xlabel('physical error rate')
    plt.ylabel('logical error rate')
    plt.title('Phenomenological Noise Model')
    for d in [3, 5, 7, 9, 11]:
        p_vec = [0.5 * (0.8 ** i) for i in range(20)]
        y_vec = []
        y_error = []
        for p in progressbar(p_vec, f"d = {d:2d}: "):
            error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, d, repeats)
            if relative_uncertainty > 0.3:
                break
            y_vec.append(error_rate)
            y_error.append(error_rate * relative_uncertainty)
        plt.errorbar(p_vec[:len(y_vec)], y_vec, y_error, label=f"d = {d}")
        print()
    plt.legend()
    # plt.show()
    plt.savefig('evaluateDecoderThresholdPhenomenological.pdf')

def evaluateDecoderThresholdCircuitLevel():
    np.random.seed(3422712)
    repeats = 10000
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.ylim(2e-3, 0.55)
    plt.xlim(1e-3, 0.5)
    plt.yscale('log')
    plt.xscale('log')
    plt.xlabel('physical error rate')
    plt.ylabel('logical error rate')
    plt.title('Circuit-Level Noise Model')
    for d in [3, 5, 7, 9, 11]:
        p_vec = [0.5 * (0.8 ** i) for i in range(25)]
        y_vec = []
        y_error = []
        for p in progressbar(p_vec, f"d = {d:2d}: "):
            error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, d, repeats, is_circuit_level=True)
            if relative_uncertainty > 0.3:
                break
            y_vec.append(error_rate)
            y_error.append(error_rate * relative_uncertainty)
        plt.errorbar(p_vec[:len(y_vec)], y_vec, y_error, label=f"d = {d}")
        print()
    plt.legend()
    plt.savefig('evaluateDecoderThresholdCircuitLevel.pdf')
    plt.show()

def evaluateDecoderThresholdCircuitLevelWithPhenomenologicalDecoder():
    np.random.seed(3422712)
    repeats = 10000
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.ylim(2e-3, 0.55)
    plt.xlim(1e-3, 0.5)
    plt.yscale('log')
    plt.xscale('log')
    plt.xlabel('physical error rate')
    plt.ylabel('logical error rate')
    plt.title('Circuit-Level Noise Model (Phenomenological Decoder)')
    for d in [3, 5, 7, 9, 11]:
        p_vec = [0.5 * (0.8 ** i) for i in range(25)]
        y_vec = []
        y_error = []
        for p in progressbar(p_vec, f"d = {d:2d}: "):
            error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, d, repeats, is_circuit_level=True
                , force_phenomenological_decoder=True)
            if relative_uncertainty > 0.3:
                break
            y_vec.append(error_rate)
            y_error.append(error_rate * relative_uncertainty)
        plt.errorbar(p_vec[:len(y_vec)], y_vec, y_error, label=f"d = {d}")
        print()
    plt.legend()
    plt.savefig('evaluateDecoderThresholdCircuitLevelWithPhenomenologicalDecoder.pdf')
    plt.show()

def evaluateLogicalErrorAccumulation():
    np.random.seed(3422712)
    repeats = 10000
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.ylim(1e-3, 0.15)
    plt.xlim(1, 200)
    plt.yscale('log')
    plt.xlabel('noisy measurements')
    plt.ylabel('logical error rate')
    plt.title('Logical Error Accumulation')
    p = 0.02
    # add d = 1: physical qubit
    for d in [1]:
        noisy_measurements_vec = [i for i in range(0, 30)]
        y_vec = []
        error_rate = p
        for noisy_measurements in noisy_measurements_vec:
            error_rate = p * (1 - error_rate) + error_rate * (1 - p)
            if error_rate > 0.15:
                break
            y_vec.append(error_rate)
        plt.errorbar(noisy_measurements_vec[:len(y_vec)], y_vec, [0 for _ in y_vec], label=f"d = {d}")
    # d = 3, 5, 7: logical qubit
    for d in [3, 5, 7]:
        noisy_measurements_vec = [i for i in range(0, 200, 2 if d == 3 else 10)]
        y_vec = []
        y_error = []
        for noisy_measurements in progressbar(noisy_measurements_vec, f"d = {d:2d}: "):
            error_rate, relative_uncertainty = evaluate_logical_error_rate(p, d, noisy_measurements, repeats)
            if error_rate > 0.1:
                break
            y_vec.append(error_rate)
            y_error.append(error_rate * relative_uncertainty)
        plt.errorbar(noisy_measurements_vec[:len(y_vec)], y_vec, y_error, label=f"d = {d}")
        print()
    plt.legend()
    plt.savefig('evaluateLogicalErrorAccumulation.pdf')
    plt.show()

def testAll():
    # public tests of 5.1
    testCommuteThroughCX00()
    testCommuteThroughCX01()
    testCommuteThroughCX10()
    testCommuteThroughCX11()
    testCommuteThroughInitialize()
    # public tests of 5.2
    testBasicSimulation()
    testNoisySimulation()
    testNoisySimulationMultipleRounds()
    testNoisySimulationAllPhenomenologicalNoise()
    testNoisySimulationAllCircuitLevelNoise()
    # public tests of 6.1 (disabled by default, enable as you need)
    # testBasicDecode()
    # testDecodeLogicalErrorRate()
    # public tests of 6.2 (disabled by default, enable as you need)
    # testDecodePhenomenological()
    # testDecodeLogicalErrorRatePhenomenological()
    # public tests of 6.3 (disabled by default, enable as you need)
    # testDecodeCircuitLevel()
    # testDecodeLogicalErrorRateCircuitLevel()
    # evaluations of 6.4 (disabled by default, enable as you need)
    # evaluateDecoderThreshold()  # this can take 1 minute to finish
    # evaluateDecoderThresholdPhenomenological()  # this can take 5 minutes to finish
    # evaluateDecoderThresholdCircuitLevel()  # this can take 15 minutes to finish
    # evaluateDecoderThresholdCircuitLevelWithPhenomenologicalDecoder()  # this can take 15 minutes to finish
    # evaluations of 6.5 (disabled by default, enable as you need)
    # evaluateLogicalErrorAccumulation()  # this can take 45 minutes to finish
    # ... You can add more or comment out tests

def main():
    requirement_A5.check()
    testAll()

if __name__ == '__main__':
    main()
